//
// Example for NV_transform_feedback extension
//
// Author: Alex V. Boreskoff <steps3d@narod.ru>
//

#include    "libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "TypeDefs.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include	"GlslProgram.h"
#include	"utils.h"
#include	"Camera.h"
#include	"VertexBuffer.h"

#define	NUM_PARTICLES	2000

Vector3D    eye   ( -7.0, 0, 2 );  		// camera position
unsigned    decalMap;                   // decal (diffuse) texture
unsigned    stoneMap, woodMap;
GLuint 		query;

float   angle = 0;
float	yaw   = 0;
float	pitch = 0;
float	roll  = 0;

Camera		camera ( eye, 0, 0, 0 );	// camera to be used

GlslProgram	program1;						// build G-buffer
GlslProgram	program2;						// test G-buffer
VertexBuffer * posBuf [2];					// buffer with vertices
VertexBuffer * velBuf [2];					// buffers with velocities
int			   bufIndex = 0;

void displayBoxes ();
void reshape      ( int w, int h );

inline	float	rnd ()
{
	return (float) rand () / (float) RAND_MAX;
}

inline	float	rnd ( float a, float b )
{
	return a + rnd () * (b - a);
}

void createBuffers ()
{
	float	posData [NUM_PARTICLES * 4];
	float	velData [NUM_PARTICLES * 4];
	int		i;
	
	for ( i = 0; i < NUM_PARTICLES; i++ )
	{
		posData [4*i + 0] = 0;
		posData [4*i + 1] = 0;
		posData [4*i + 2] = 5;
		posData [4*i + 3] = 1;
		velData [4*i + 0] = rnd (-1,1)*0.4;
		velData [4*i + 1] = rnd (-1,1)*0.4;
		velData [4*i + 2] = rnd (-1,1)*0.4;
		velData [4*i + 3] = 0;
	}
	
	for ( i = 0; i < 2; i++ )
	{
		posBuf [i] = new VertexBuffer ();
		velBuf [i] = new VertexBuffer ();
		
		posBuf [i] -> bind    ( GL_ARRAY_BUFFER_ARB );
		posBuf [i] -> setData ( NUM_PARTICLES * 4 * sizeof ( float ), posData, GL_STREAM_COPY );
		posBuf [i] -> unbind  ();

		velBuf [i] -> bind    ( GL_ARRAY_BUFFER_ARB );
		velBuf [i] -> setData ( NUM_PARTICLES * 4 * sizeof ( float ), velData, GL_STREAM_COPY );
		velBuf [i] -> unbind  ();
	}
}

void displayPoints ( int index, GlslProgram& program )
{
    float quadratic [] =  { 1.0f, 0.0f, 0.01f };

	glEnable      ( GL_TEXTURE_2D );
	glBindTexture ( GL_TEXTURE_2D, decalMap );
	glDepthMask   ( GL_FALSE );
	glEnable      ( GL_BLEND );
	glBlendFunc   ( GL_ONE, GL_ONE );
    glPointSize   ( 5 );
//	glColor4f ( 1, 1, 1, 1 );
	
    glEnable              ( GL_POINT_SPRITE_ARB );
    glPointParameterfvARB ( GL_POINT_DISTANCE_ATTENUATION_ARB, quadratic );
    glPointParameterfARB  ( GL_POINT_FADE_THRESHOLD_SIZE_ARB,  20.0f );
	glEnable              ( GL_VERTEX_PROGRAM_POINT_SIZE_ARB );
    glTexEnvf             ( GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE );

	glPushClientAttrib  ( GL_CLIENT_VERTEX_ARRAY_BIT );

												// setup vertices array
	posBuf [index] -> bind ( GL_ARRAY_BUFFER_ARB );
	glEnableClientState ( GL_VERTEX_ARRAY );
    glVertexPointer     ( 4, GL_FLOAT, 0, (void *)0 );

	posBuf [index] -> unbind ();
	velBuf [index] -> bind ( GL_ARRAY_BUFFER_ARB );
												// setup texcoord0 array
	glEnableClientState ( GL_TEXTURE_COORD_ARRAY );
	glTexCoordPointer	( 4, GL_FLOAT, 0, (void *)0 );

	velBuf [index] -> unbind ();
												// setup l array
	glDrawArrays ( GL_POINTS, 0, NUM_PARTICLES );

	glPopClientAttrib ();
	glDepthMask       ( GL_TRUE );
	glDisable         ( GL_BLEND );
}

void display ()
{
	unsigned primitivesWritten;
	
    glClear                       ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	program1.bind                 ();
	glBindBufferOffsetNV          ( GL_TRANSFORM_FEEDBACK_BUFFER_NV, 0, posBuf [bufIndex ^ 1] -> getId (), 0 );
	glBindBufferOffsetNV          ( GL_TRANSFORM_FEEDBACK_BUFFER_NV, 1, velBuf [bufIndex ^ 1] -> getId (), 0 );
	glBeginTransformFeedbackNV    ( GL_POINTS );
    glEnable                      ( GL_RASTERIZER_DISCARD_NV );    // disable rasterization
    glBeginQueryARB               ( GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN_NV, query );
	displayPoints                 ( bufIndex, program1 );
    glDisable                     ( GL_RASTERIZER_DISCARD_NV );
    glEndQueryARB                 ( GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN_NV );
	glEndTransformFeedbackNV      ();
	program1.unbind               ();

    glGetQueryObjectuivARB ( query, GL_QUERY_RESULT, &primitivesWritten );
/*	
printf ( "Prim. written: %d\n", primitivesWritten );

float	buf [8], buf2 [8];

posBuf [bufIndex ^ 1] -> bind (GL_ARRAY_BUFFER_ARB);
posBuf [bufIndex ^ 1] -> getSubData ( 0, 4*1*sizeof ( float ), buf );
posBuf [bufIndex ^ 1] -> unbind ();

glutReportErrors ();
printf ( "Pos: %f %f %f %f\n", buf [0], buf [1], buf [2], buf [3] );

velBuf [bufIndex ^ 1] -> bind (GL_ARRAY_BUFFER_ARB);
velBuf [bufIndex ^ 1] -> getSubData ( 0, 4*1*sizeof ( float ), buf2 );
velBuf [bufIndex ^ 1] -> unbind ();

printf ( "Vel: %f %f %f %f\n", buf2 [0], buf2 [1], buf2 [2], buf2 [3] );

glutReportErrors ();
*/
	camera.apply ();

// draw the reflectors

	glEnable      ( GL_TEXTURE_2D );
	glBindTexture ( GL_TEXTURE_2D, stoneMap );
	glBegin       ( GL_QUADS );
		glTexCoord2f ( 0, 0 );
		glVertex3f   ( -10.0, -10.0, -2.0  );
		glTexCoord2f ( 10, 0 );
		glVertex3f   (  10.0, -10.0, -2.0  );
		glTexCoord2f ( 10, 10 );
		glVertex3f   (  10.0,  10.0, -2.0  );
		glTexCoord2f ( 0, 10 );
		glVertex3f   (  -10.0,  10.0, -2.0  );
	glEnd         ();
		
	glEnable      ( GL_TEXTURE_2D );
	glBindTexture ( GL_TEXTURE_2D, woodMap );
	glBegin       ( GL_QUADS );
		glTexCoord2f ( 0, 0 );
		glVertex3f ( 0.0, 0.0, -1.5 );
		glTexCoord2f ( 1, 0 );
		glVertex3f ( 2.0, 0.0, -1.5 );
		glTexCoord2f ( 1, 1 );
		glVertex3f ( 2.0, 2.0, -1.5 );
		glTexCoord2f ( 0, 1 );
		glVertex3f ( 0.0, 2.0, -1.5 );
		
		glTexCoord2f ( 0, 0 );
		glVertex3f ( -2.0,  -2.0, -1.0 );
		glTexCoord2f ( 1, 0 );
		glVertex3f (  0.0,  -2.0, -1.0 );
		glTexCoord2f ( 1, 1 );
		glVertex3f (  0.0,   0.0, -1.0 );
		glTexCoord2f ( 0, 1 );
		glVertex3f ( -2.0,   0.0, -1.0 );
	glEnd         ();
	
	program2.bind   ();
	displayPoints   ( bufIndex ^ 1, program2 );
	program2.unbind ();
	
	bufIndex ^= 1;
	
    glutSwapBuffers ();
}

void reshape ( int w, int h )
{
	camera.setViewSize ( w, h, 60 );
	camera.apply       ();
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );
    else
   	if ( key == 'w' || key == 'W' )
   		camera.moveBy ( camera.getViewDir () * 0.2 );
   	else
   	if ( key == 'x' || key == 'X' )
   		camera.moveBy ( -camera.getViewDir () * 0.2 );
   	else
   	if ( key == 'a' || key == 'A' )
   		camera.moveBy ( -camera.getSideDir () * 0.2 );
   	else
   	if ( key == 'd' || key == 'D' )
   		camera.moveBy ( camera.getSideDir () * 0.2 );

   	glutPostRedisplay ();
}

void    specialKey ( int key, int x, int y )
{
    if ( key == GLUT_KEY_UP )
        yaw += M_PI / 90;
    else
    if ( key == GLUT_KEY_DOWN )
        yaw -= M_PI / 90;
	else
    if ( key == GLUT_KEY_RIGHT )
        roll += M_PI / 90;
    else
    if ( key == GLUT_KEY_LEFT )
        roll -= M_PI / 90;

	camera.setEulerAngles ( yaw, pitch, roll );

    glutPostRedisplay ();
}

void	mouseFunc ( int x, int y )
{
	static	int	lastX = -1;
	static	int	lastY = -1;

	if ( lastX == -1 )				// not initialized
	{
		lastX = x;
		lastY = y;
	}

	yaw  -= (y - lastY) * 0.02;
	roll += (x - lastX) * 0.02;

	lastX = x;
	lastY = y;

	camera.setEulerAngles ( yaw, pitch, roll );

	glutPostRedisplay ();
}

void    animate ()
{
	static	float	lastTime = 0.0;
	float			time     = 0.001f * glutGet ( GLUT_ELAPSED_TIME );
	float			dt       = time - lastTime;
	
    lastTime = time;

	program1.bind   ();
	program1.setUniformFloat ( "dt", dt );
	program1.unbind ();
	
    glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
    glutInitWindowSize  ( 640, 480 );


                                // create window
    glutCreateWindow ( "OpenGL NV_transform_feedback example" );

                                // register handlers
    glutDisplayFunc       ( display    );
    glutReshapeFunc       ( reshape    );
    glutKeyboardFunc      ( key        );
    glutSpecialFunc       ( specialKey );
    glutPassiveMotionFunc ( mouseFunc  );
    glutIdleFunc          ( animate    );

    init           ();
    initExtensions ();

    assertExtensionsSupported ( "NV_transform_feedback" );

	decalMap = createTexture2D ( true, "../../Textures/Fire.bmp" );
	stoneMap = createTexture2D ( true, "../../Textures/16.jpg"   );
	woodMap  = createTexture2D ( true, "../../Textures/oak.bmp"  );

	createBuffers   ();
    glGenQueriesARB ( 1, &query );
	
	if ( !program1.loadShaders ( "transform.vsh", "transform.fsh" ) )
	{
		printf ( "Error loading transform shaders:\n%s\n", program1.getLog ().c_str () );

		return 3;
	}
	
	if ( !program2.loadShaders ( "particles.vsh", "particles.fsh" ) )
	{
		printf ( "Error loading drawing shaders:\n%s\n", program2.getLog ().c_str () );

		return 3;
	}
	
									// register newPos and newVal as active
	program1.addActiveVaryings ( "newPos" );
	program1.addActiveVaryings ( "newVel" );
	program1.relink            ();			// relink required
	
	int	locs [2];
	
	locs [0] = glGetVaryingLocationNV ( program1.getProgram (), "newPos" );
	locs [1] = glGetVaryingLocationNV ( program1.getProgram (), "newVel" );

	program1.bind                 ();
	glTransformFeedbackVaryingsNV ( program1.getProgram (), 2, locs, GL_SEPARATE_ATTRIBS_NV );
	program1.unbind               ();

	program2.bind        ();
	program2.setTexture  ( "decalMap", 0 );
	program2.unbind      ();

	camera.setRightHanded ( false );
	
    glutMainLoop ();

    return 0;
}
